import os
from pathlib import Path

WEBAPP_DIR = Path(os.path.dirname(__file__))
TEMPLATE_DIR = WEBAPP_DIR / "templates"
TEMPLATE_DICT = {
    "Binary classification": {
        "1. Dataset preparation": TEMPLATE_DIR
        / "binary_classification"
        / "dataset_preparation.py",
        "2. Preprocessing": TEMPLATE_DIR
        / "binary_classification"
        / "preprocessing.py",
        "3. Feature extraction": TEMPLATE_DIR
        / "binary_classification"
        / "feature_extraction.py",
        "4. Training": TEMPLATE_DIR / "binary_classification" / "training.py",
        "5. Evaluation": TEMPLATE_DIR
        / "binary_classification"
        / "evaluation.py",
        "6. Inference": TEMPLATE_DIR
        / "binary_classification"
        / "inference.py",
    },
    "Radiomics maps": {
        "Create maps": TEMPLATE_DIR / "radiomics_maps" / "create_maps.py",
    },
    "Segmentation": {
        "Generate code for nnUNet": TEMPLATE_DIR
        / "segmentation"
        / "segmentation.py",
    },
}

IS_DEMO = False
if "INPUT_DIR" not in os.environ:
    IS_DEMO = True



import logging

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import auc, precision_recall_curve, roc_curve

from autorad.visualization.plotly_utils import plot_roc_curve, plot_waterfall

from .utils import get_sensitivity_specificity, get_youden_threshold

log = logging.getLogger(__name__)


class SimpleEvaluator:
    def __init__(self, y_true, y_pred_proba):
        self.y_true = y_true
        self.y_pred_proba = y_pred_proba
        self.threshold = get_youden_threshold(self.y_true, self.y_pred_proba)
        self.y_pred = self.y_pred_proba > self.threshold

    def plot_roc_curve(self):
        fig = plot_roc_curve(self.y_true, self.y_pred_proba)

        return fig

    def plot_precision_recall_curve(self, title=None):
        """
        Plot the precision recall curve.
        """
        fig, ax = plt.subplots(figsize=(10, 10))
        precision, recall, _ = precision_recall_curve(
            self.y_true, self.y_pred_proba
        )
        auc_score = np.round(auc(recall, precision), 3)
        ax.plot(
            recall,
            precision,
            lw=2,
            alpha=0.8,
            label=f"AUC={auc_score}",
        )
        ax.set_xlabel("Recall", fontsize=40)
        ax.set_ylabel("Precision", fontsize=40)
        ax.legend(loc="lower left", fontsize=40)
        if title:
            ax.set_title(title)
        else:
            ax.set_title("Precision-Recall Curve", fontsize=40)
        fig.tight_layout()

        return fig

    def plot_waterfall(self):
        fig = plot_waterfall(self.y_true, self.y_pred_proba, self.threshold)
        return fig

    def plot_optimal_point_test(self, ax):
        fpr, tpr, _ = roc_curve(self.y_true, self.y_pred)
        fpr_point = fpr[1]
        tpr_point = tpr[1]
        sens, spec = get_sensitivity_specificity(
            self.y_true, self.y_pred, self.threshold
        )
        point_label = f"Sensitivity = {sens}, Specificity = {spec}"
        ax.plot(
            [fpr_point],
            [tpr_point],
            marker="h",
            markersize=10,
            color="black",
            label=point_label,
        )
        return self



import logging
import pandas as pd
import os
from pathlib import Path
from typing import List, Optional, Sequence

from joblib import Parallel, delayed

from autorad.config.type_definitions import PathLike
from autorad.data.dataset import ImageDataset
from autorad.utils import spatial

log = logging.getLogger(__name__)


def generate_border_masks(
    dataset: ImageDataset,
    margin_in_mm: float | Sequence[float],
    output_dir: PathLike,
    n_jobs: int = -1,
):
    """
    Generate a border mask (= mask with given margin around the original ROI)
    for each mask in the dataset.
    Returns a DataFrame extending ImageDataset.df with the additional column
    "dilated_mask_path_<margin_in_mm>".
    """
    if n_jobs == -1:
        n_jobs = os.cpu_count()
    output_paths = [
        os.path.join(output_dir, f"{id_}_border_mask.nii.gz")
        for id_ in dataset.ids
    ]
    if n_jobs > 1:
        with Parallel(n_jobs) as parallel:
            parallel(
                delayed(spatial.get_border_outside_mask_mm)(
                    mask_path=mask_path,
                    margin=margin_in_mm,
                    output_path=output_path,
                )
                for mask_path, output_path in zip(
                    dataset.mask_paths, output_paths
                )
            )
    else:
        for mask_path, output_path in zip(dataset.mask_paths, output_paths):
            spatial.get_border_outside_mask_mm(
                mask_path,
                margin_in_mm,
                output_path,
            )
    result_df = dataset.df.copy()
    result_df[f"border_mask_path_{margin_in_mm}mm"] = output_paths

    return result_df


def get_paths_with_separate_folder_per_case_loose(
    data_dir: PathLike,
    image_stem: str = "image",
    mask_stem: str = "segmentation",
    relative: bool = False,
    ) -> pd.DataFrame:
    ids, image_paths, mask_paths = [], [], []
    for id_ in os.listdir(data_dir):
        id_dir = os.path.join(data_dir, id_)
        if not os.path.isdir(id_dir):
            continue
        case_image_paths = Path(id_dir).glob("*.nii.gz")
        case_image_paths = [
            p for p in case_image_paths if image_stem in p.name
        ]
        if not len(case_image_paths) == 1:
            log.error(
                f"Expected 1, found {len(image_paths)} images for ID={id_}"
            )
            continue
        image_path = case_image_paths[0]
        for fname in os.listdir(id_dir):
            if mask_stem in fname:
                mask_path = os.path.join(id_dir, fname)

                ids.append(id_)
                image_paths.append(str(image_path))
                mask_paths.append(str(mask_path))
    if relative:
        image_paths = make_relative(image_paths, data_dir)
        mask_paths = make_relative(mask_paths, data_dir)
    path_df = paths_to_df(ids, image_paths, mask_paths)
    
    return path_df


def get_paths_with_separate_folder_per_case(
    data_dir: PathLike,
    image_stem: str = "image",
    mask_stem: str = "segmentation",
    relative: bool = False,
    ) -> pd.DataFrame:
    ids, image_paths, mask_paths = [], [], []
    for id_ in os.listdir(data_dir):
        id_dir = os.path.join(data_dir, id_)
        if not os.path.isdir(id_dir):
            continue
        image_path = os.path.join(id_dir, f"{image_stem}.nii.gz")
        mask_path = os.path.join(id_dir, f"{mask_stem}.nii.gz")
        if not os.path.exists(image_path):
            log.error(f"Image for ID={id_} does not exist ({image_path})")
            continue
        if not os.path.exists(mask_path):
            log.error(f"Mask for ID={id_} does not exist ({mask_path})")
            continue
        ids.append(id_)
        image_paths.append(image_path)
        mask_paths.append(mask_path)
    if relative:
        image_paths = make_relative(image_paths, data_dir)
        mask_paths = make_relative(mask_paths, data_dir)
    path_df = paths_to_df(ids, image_paths, mask_paths)

    return path_df


def get_paths_with_separate_image_seg_folders(
    image_dir: PathLike,
    mask_dir: PathLike,
    relative_to: Optional[PathLike] = None,
    ) -> pd.DataFrame:
    """Get paths of images and segmentations when all images are in one folder
    and all segmentations are in another folder. It assumes the file names
    for images and segmentation are the same, or image filenames may have optional
    suffix _XXXX, stemming from nnUNet.

    Args:
        image_dir (PathLike): Path to the folder with the images.
        mask_dir (PathLike): Path to the folder with the segmentations.
        relative_to (Optional[PathLike], optional): put a root directory to get relative paths.
            If None, return absolute paths.

    Returns:
        tuple[List[str], List[str], List[str]]: IDs, image paths, and mask paths
    """
    images = list(Path(image_dir).glob("*.nii.gz"))
    masks = list(Path(mask_dir).glob("*.nii.gz"))
    mask_ids = [p.stem.split(".")[0] for p in masks]
    mask_dict = {id_: str(p) for id_, p in zip(mask_ids, masks)}
    ids_matched, images_matched, masks_matched = [], [], []
    for image in images:
        id_ = image.stem.split(".")[0]
        if id_.endswith("_0000"):
            id_ = id_[:-5]
        if id_ not in mask_ids:
            log.error(f"No mask found for ID={id_}")
        else:
            ids_matched.append(id_)
            images_matched.append(str(image))
            masks_matched.append(str(mask_dict[id_]))
    if relative_to is not None:
        images_matched = make_relative(images_matched, relative_to)
        masks_matched = make_relative(masks_matched, relative_to)

    path_df = paths_to_df(ids_matched, images_matched, masks_matched)

    return path_df

def paths_to_df(ids, image_paths, mask_paths):
    df = pd.DataFrame({
            "ID": ids,
            "image_path": image_paths,
            "segmentation_path": mask_paths
        })
    return df

def make_relative(paths: List[PathLike], root_dir):
    return [os.path.relpath(p, root_dir) for p in paths]



from __future__ import annotations

from pathlib import Path

import git
import pandas as pd
from monai.apps.datasets import DecathlonDataset, MedNISTDataset

from autorad.config import config
from autorad.config.type_definitions import PathLike
from autorad.data.dataset import ImageDataset


def load_mednist_dataset(root_dir: PathLike) -> ImageDataset:
    """
    Downlad, load MONAI MedNISTDataset
    and convert it to ImageDataset.
    """
    mednist_dataset = MedNISTDataset(
        root_dir=root_dir, section="training", download=True
    )
    data_list = mednist_dataset.data
    image_paths = [elem["image"] for elem in data_list]
    labels = [elem["label"] for elem in data_list]
    class_names = [elem["class_name"] for elem in data_list]
    ids = range(len(image_paths))
    df = pd.DataFrame(
        {
            "id": ids,
            "image_path": image_paths,
            "label": labels,
            "class_name": class_names,
        }
    )
    return ImageDataset(
        df=df, image_colname="image_path", mask_colname="mask_path"
    )


def convert_decathlon_dataset(decathlon_dataset: DecathlonDataset):
    """
    Convert the MONAI DecathlonDataset into a
    autorad.data.dataset.ImageDataset.
    """
    data_list = decathlon_dataset.data
    image_paths = [elem["image"] for elem in data_list]
    mask_paths = [elem["label"] for elem in data_list]
    ids = range(len(image_paths))
    df = pd.DataFrame(
        {
            "id": ids,
            "image_path": image_paths,
            "mask_path": mask_paths,
        }
    )
    return ImageDataset(
        df=df, image_colname="image_path", mask_colname="mask_path"
    )


def load_prostatex(root_dir: PathLike | None = None) -> ImageDataset:
    if root_dir is None:
        root_dir = Path(config.TEST_DATA_DIR) / "datasets" / "PROSTATEx"
    if not Path(root_dir).exists() or Path(root_dir).is_file():
        git.Repo.clone_from(
            "https://github.com/rcuocolo/PROSTATEx_masks/", root_dir
        )


if __name__ == "__main__":
    load_prostatex()



from pathlib import Path

import pandas as pd
import streamlit as st

from autorad.external.download_WORC import download_WORCDatabase
from autorad.utils import preprocessing
from autorad.webapp import utils, validation_utils


def download_example_dataset(input_dir):
    with st.expander("You have no data? Then download an example dataset 🫁"):
        col1, col2 = st.columns(2)
        with col1:
            n_subjects = st.slider(
                "Number of cases", min_value=5, max_value=100, value=10
            )
        with col2:
            st.text_input("Where to save the example dataset", value=input_dir)
            if not Path(input_dir).exists():
                st.error("Result directory does not exist.")
            dataset_save_dir = Path(input_dir) / "WORC_desmoid"
        if st.button("Download example dataset"):
            with st.spinner("Downloading dataset..."):
                download_WORCDatabase(
                    dataset="Desmoid",
                    data_folder=dataset_save_dir,
                    n_subjects=n_subjects,
                )
            st.success(
                f"Downloaded example dataset to {dataset_save_dir}.\n"
                f"You can use this path to generate the table with paths below."
            )


def show():
    input_dir = utils.get_input_dir()
    download_example_dataset(input_dir)

    st.write(
        """
    #####
    Before you proceed with the next tasks you need to create a .csv table.
    The table should contain the following information for each case:
    - **ID** (has to be unique)
    - Path to the **image**
    - Path to the **segmentation**
    - Optionally: **Label** (0 or 1)

    Example:
    """
    )
    example_path_df = pd.read_csv(
        Path(__file__).parents[2] / "paths_example.csv"
    )
    st.write(example_path_df)
    st.write(
        """
    #####
    You can create the table manually (e.g. in Excel) or, if your dataset
    follows the specified structure, you can **generate it below** ⬇
    """
    )
    dir_structures = {
        "Separate folder for each case": """
            Expected format:
            ```
            <dataset folder>
            +-- case1
            |   +-- image.nii.gz
            |   +-- segmentation.nii.gz
            +-- case2
            |   +-- image.nii.gz
            |   +-- segmentation.nii.gz
            ...
            ```
            """,
        "Separate folder for each case with multiple segmentations per image": """
            Expected format:
            ```
            <dataset folder>
            +-- case1
            |   +-- image.nii.gz
            |   +-- segmentation1.nii.gz
            |   +-- segmentation002.nii.gz
            |   +-- segmentation_03.nii.gz
            ...
            +-- case2
            |   +-- image.nii.gz
            |   +-- segmentation_first.nii.gz
            |   +-- segmentation_second.nii.gz
            ...
            ```
            """,
        "One folder for all the images, and one for the segmentations": """
            Expected format:
            ```
            <dataset folder>
            +-- images
            |   +-- 1.nii.gz
            |   +-- 2.nii.gz
            |   +-- ...
            +-- segmentations
            |   +-- 1.nii.gz
            |   +-- 2.nii.gz
            |   +-- ...
            ...
            ```
            """,
    }
    formats = list(dir_structures.keys())
    col1, col2 = st.columns(2)
    with col1:
        format = st.radio("Dataset structure:", formats)
    with col2:
        st.write(dir_structures[format])
    data_dir = st.text_input(
        "Enter path to the root directory of your dataset:",
        value=input_dir,
    )
    if not data_dir:
        st.stop()
    data_dir = Path(data_dir).absolute()
    validation_utils.check_if_dir_exists(data_dir)
    col1, col2 = st.columns(2)

    if format == formats[0] or format == formats[1]:
        with col1:
            image_stem = st.text_input(
                "How to identify image files? What root does every filename contain?",
                value="image",
            )
        with col2:
            mask_stem = st.text_input(
                "How to identify segmentation files? What root does every filename contain?",
                value="segmentation",
            )
    if format == formats[0]:
        paths_df = preprocessing.get_paths_with_separate_folder_per_case(
            data_dir=data_dir,
            image_stem=image_stem,
            mask_stem=mask_stem,
        )
    elif format == formats[1]:
        paths_df = preprocessing.get_paths_with_separate_folder_per_case_loose(
            data_dir=data_dir,
            image_stem=image_stem,
            mask_stem=mask_stem,
        )
    else:
        with col1:
            image_dir = st.text_input(
                "Enter path to the folder with images:",
                value=data_dir / "images",
            )
        with col2:
            segmentation_dir = st.text_input(
                "Enter path to the folder with segmentations:",
                value=data_dir / "segmentations",
            )
        paths_df = preprocessing.get_paths_with_separate_image_seg_folders(
            image_dir=image_dir,
            mask_dir=segmentation_dir,
        )
    if paths_df.empty:
        st.warning(
            "Looks like no files were found."
            " Please check the dataset root directory and file names."
        )
        st.stop()
    else:
        st.success(f"Found {len(paths_df)} cases.")
        st.dataframe(paths_df)
        utils.save_table_in_result_dir(paths_df, "paths.csv")


if __name__ == "__main__":
    show()



def get_organ_names(models_metadata: dict):
    organs = []
    for model_meta in models_metadata.values():
        for organ_name in model_meta["labels"].values():
            if organ_name not in organs:
                organs.append(organ_name)
    return sorted(organs)


def get_region_names(models_metadata: dict):
    regions = [model_meta["region"] for model_meta in models_metadata.values()]
    regions = sorted(list(set(regions)))
    return regions


def filter_models_metadata(
    models_metadata: dict, modality: str, region_name: str
):
    matching_models_metadata = {}
    for model_name, model_meta in models_metadata.items():
        if (
            model_meta["modality"] == modality
            and model_meta["region"] == region_name
        ):
            matching_models_metadata[model_name] = model_meta
    return matching_models_metadata


def filter_models_metadata_by_organ(models_metadata: dict, organ: str):
    matching_models_metadata = {}
    for model_name, model_meta in models_metadata.items():
        if organ in model_meta["labels"].values():
            matching_models_metadata[model_name] = model_meta
    return matching_models_metadata


def get_organ_label(model_metadata: dict, organ: str):
    for label in model_metadata["labels"]:
        if model_metadata["labels"][label] == organ:
            return int(label)



# def plot_feature_importance(
#     dataset: FeatureDataset, model: MLClassifier, ax: plt.Axes
# ):
#     """
#     Plot importance of features for a single model
#     Args:
#         model [MLClassifier] - classifier
#         ax (optional) - pyplot axes object
#     """
#     model_name = model.name
#     try:
#         importances = model.feature_importance()
#         importance_df = pd.DataFrame(
#             {
#                 "feature": dataset.features,
#                 "importance": importances,
#             }
#         )
#         sns.barplot(x="feature", y="importance", data=importance_df, ax=ax)
#         ax.tick_params(axis="both", labelsize="x-small")
#         ax.set_ylabel("Feature importance")
#         ax.set_title(model_name)
#         ax.set_xticklabels(ax.get_xticklabels(), rotation=30, ha="right")
#     except Exception:
#         log.error(f"For {model_name} feature importance cannot be calculated.")


# def plot_feature_importance_all(
#     dataset: FeatureDataset,
#     models: list[MLClassifier],
#     result_dir: PathLike,
#     title: Optional[str] = None,
# ):
#     """
#     Plot the feature importance for all models.
#     """
#     nrows, ncols, figsize = get_subplots_dimensions(len(models))
#     fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
#     for i, model in enumerate(models):
#         ax = fig.axes[i]
#         plot_feature_importance(dataset, model, ax=ax)
#     if title:
#         fig.suptitle(title)
#     else:
#         fig.suptitle(f"Feature Importance for {dataset.task_name}")
#     fig.tight_layout()
#     fig.savefig(
#         Path(result_dir) / "feature_importance.png",
#         bbox_inches="tight",
#         dpi=100,
#     )
#     plt.show()


# def plot_lofo_importance(dataset: FeatureDataset, model: MLClassifier):
#     lofo_dataset = lofo.Dataset(
#         df=dataset.df,
#         target=dataset.target,
#         features=dataset.selected_features,
#     )
#     lofo_imp = lofo.LOFOImportance(
#         lofo_dataset, model=model.classifier, scoring="neg_mean_squared_error"
#     )
#     importance_df = lofo_imp.get_importance()
#     lofo.plot_importance(importance_df, figsize=(12, 12))
#     plt.tight_layout()
#     plt.show()


def common_roc_settings(ax, fontsize=12):
    ax.plot([0, 1], [0, 1], linestyle="--", lw=2, color="black", alpha=0.8)
    ax.set_xlabel("False Positive Rate", fontsize=fontsize)
    ax.set_ylabel("True Positive Rate", fontsize=fontsize)
    ax.legend(loc="lower right", fontsize="large")


def get_subplots_dimensions(n_plots):
    """
    For given number of plots returns the 'optimal' rows x columns distribution
    of subplots and figure size.
    Args:
        n_plots [int] - number of subplots to be includeed in the plot
    Returns:
        nrows [int] - suggested number of rows ncols [int] - suggested number of
        columns figsize [tuple[int, int]] - suggested figsize
    """
    if n_plots == 1:
        nrows = 1
        ncols = 1
        figsize = (12, 7)
    elif n_plots == 2:
        nrows = 1
        ncols = 2
        figsize = (13, 6)
    elif n_plots == 3:
        nrows = 1
        ncols = 3
        figsize = (20, 5)
    elif n_plots == 4:
        nrows = 2
        ncols = 2
        figsize = (14, 8)
    elif n_plots in [5, 6]:
        nrows = 2
        ncols = 3
        figsize = (20, 9)
    elif n_plots == 9:
        nrows = 3
        ncols = 3
        figsize = (18, 12)
    elif n_plots == 10:
        nrows = 2
        ncols = 5
        figsize = (20, 7)
    elif n_plots > 4:
        nrows = n_plots // 4 + 1
        ncols = 4
        figsize = (20, 7 + 5 * nrows)
    else:
        raise ValueError("Invalid number of plots")

    return nrows, ncols, figsize



import importlib.util
import os
import tempfile
import zipfile
from pathlib import Path
from shutil import copy

import git
import pandas as pd
import streamlit as st


def save_table(df, save_path):
    df.to_csv(save_path, index=False)
    st.success(f"Done! Table saved ({save_path})")


def save_table_in_result_dir(df, filename, button=True):
    if button:
        saved = st.button("Looks good? Save the table ⬇️")
    else:
        saved = True
    if saved:
        result_dir = get_result_dir()
        out_path = Path(result_dir) / filename
        df.to_csv(out_path, index=False)
        st.success(f"Done! Table saved in your result directory ({out_path})")


def get_input_dir():
    if "AUTORAD_INPUT_DIR" in os.environ:
        return os.environ["AUTORAD_INPUT_DIR"]
    else:
        st.error(
            "Oops, input directory not set! Go to the config page and set it."
        )
        st.stop()


def get_result_dir():
    if "AUTORAD_RESULT_DIR" in os.environ:
        return os.environ["AUTORAD_RESULT_DIR"]
    else:
        st.error(
            "Oops, result directory not set! Go to the config page and set it."
        )
        st.stop()


def get_env_var(name, show_error=False):
    try:
        return os.environ[name]
    except KeyError:
        message = f"{name} not set"
        if show_error:
            st.error(message)
            st.stop()
        return message


def zip_directory(folder_path, zip_path):
    with zipfile.ZipFile(zip_path, mode="w") as zipf:
        len_dir_path = len(folder_path)
        for root, _, files in os.walk(folder_path):
            for file in files:
                file_path = os.path.join(root, file)
                zipf.write(file_path, file_path[len_dir_path:])


# from https://github1s.com/jrieke/traingenerator
#      /blob/HEAD/app/utils.py#L1-L177
def import_from_file(module_name: str, filepath: str):
    """
    Imports a module from file.

    Args:
        module_name (str): Assigned to the module's __name__ parameter
        (does not influence how the module is named outside of this function)
        filepath (str): Path to the .py file

    Returns:
        The module
    """
    spec = importlib.util.spec_from_file_location(module_name, filepath)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def upload_file(label):
    uploaded_file = st.file_uploader(label)
    if uploaded_file is None:
        st.stop()
    return uploaded_file


def load_df(label):
    """ """
    uploaded_file = upload_file(label)
    df = pd.read_csv(uploaded_file)
    return df


def load_test_data(out_dir):
    """
    Loads test data to the `out_dir` directory
    """
    test_out_dir = Path(out_dir) / "test_data"
    test_out_dir.mkdir(exist_ok=True)
    if not dir_nonempty(test_out_dir):
        repo_address = "https://github.com/pwoznicki/AutoRadiomics.git"
        with tempfile.TemporaryDirectory() as tmp_dir:
            git.Git(tmp_dir).clone(repo_address)
            tmp_dir = Path(tmp_dir)
            for fpath in tmp_dir.rglob("*.nii.gz"):
                copy(fpath, test_out_dir)


def dir_nonempty(dir_path):
    path = Path(dir_path)
    if not path.is_dir():
        return False
    has_next = next(path.iterdir(), None)
    if has_next is None:
        return False
    return True



import numpy as np
import pandas as pd
from sklearn.metrics import auc, confusion_matrix, roc_curve


def get_youden_threshold(y_true, y_score):
    """
    Returns optimal threshold that maximizes the
    difference between TPR and FPR
    """
    fpr, tpr, thresholds = roc_curve(y_true, y_score)
    youden_index = [tpr[i] - fpr[i] for i in range(len(tpr))]
    youden_argmax = np.argmax(youden_index)
    you_thr = thresholds[youden_argmax]

    return you_thr


def get_optimal_threshold(y_test, y_score):
    fpr, tpr, thresholds = roc_curve(y_test, y_score)
    i = np.arange(len(tpr))
    roc = pd.DataFrame(
        {
            "tf": pd.Series(tpr - (1 - fpr), index=i),
            "threshold": pd.Series(thresholds, index=i),
        }
    )
    roc_t = roc.ix[(roc.tf - 0).abs().argsort()[:1]]
    return list(roc_t["threshold"])


def get_sensitivity_specificity(y_test, y_pred_proba, threshold):
    y_pred = np.where(y_pred_proba > threshold, 1, 0)
    tn, fp, fn, tp = confusion_matrix(y_test, y_pred).ravel()
    sensitivity = tp / (tp + fn)
    specificity = tn / (tn + fp)
    sensitivity = np.round(sensitivity, 3)
    specificity = np.round(specificity, 3)
    return sensitivity, specificity


def get_fpr_tpr_auc(y_true, y_pred):
    fpr, tpr, _ = roc_curve(y_true, y_pred)
    roc_auc = np.round(auc(fpr, tpr), 3)
    return fpr, tpr, roc_auc



import functools

import monai
import numpy as np
import torch
from medpy import metric

from autorad.utils import statistics


def assert_shape(test, reference):
    assert test.shape == reference.shape, "Shape mismatch: {} and {}".format(
        test.shape, reference.shape
    )


class ConfusionMatrix:
    def __init__(self, test=None, reference=None):

        self.tp = None
        self.fp = None
        self.tn = None
        self.fn = None
        self.size = None
        self.reference_empty = None
        self.reference_full = None
        self.test_empty = None
        self.test_full = None
        self.set_reference(reference)
        self.set_test(test)

    def set_test(self, test):

        self.test = test
        self.reset()

    def set_reference(self, reference):

        self.reference = reference
        self.reset()

    def reset(self):

        self.tp = None
        self.fp = None
        self.tn = None
        self.fn = None
        self.size = None
        self.test_empty = None
        self.test_full = None
        self.reference_empty = None
        self.reference_full = None

    def compute(self):

        if self.test is None or self.reference is None:
            raise ValueError(
                "'test' and 'reference' must both be set to compute confusion matrix."
            )

        assert_shape(self.test, self.reference)

        self.tp = int(((self.test != 0) * (self.reference != 0)).sum())
        self.fp = int(((self.test != 0) * (self.reference == 0)).sum())
        self.tn = int(((self.test == 0) * (self.reference == 0)).sum())
        self.fn = int(((self.test == 0) * (self.reference != 0)).sum())
        self.size = int(np.prod(self.reference.shape, dtype=np.int64))
        self.test_empty = not np.any(self.test)
        self.test_full = np.all(self.test)
        self.reference_empty = not np.any(self.reference)
        self.reference_full = np.all(self.reference)

    def get_matrix(self):

        for entry in (self.tp, self.fp, self.tn, self.fn):
            if entry is None:
                self.compute()
                break

        return self.tp, self.fp, self.tn, self.fn

    def get_size(self):

        if self.size is None:
            self.compute()
        return self.size

    def get_existence(self):

        for case in (
            self.test_empty,
            self.test_full,
            self.reference_empty,
            self.reference_full,
        ):
            if case is None:
                self.compute()
                break

        return (
            self.test_empty,
            self.test_full,
            self.reference_empty,
            self.reference_full,
        )


# Borrowed from https://github.com/MIC-DKFZ/nnUNet/blob/HEAD/nnunet/evaluation/metrics.py#L141-L175
def specificity(
    test=None,
    reference=None,
    confusion_matrix=None,
    nan_for_nonexisting=True,
    **kwargs
):
    """TN / (TN + FP)"""

    if confusion_matrix is None:
        confusion_matrix = ConfusionMatrix(test, reference)

    tp, fp, tn, fn = confusion_matrix.get_matrix()
    (
        test_empty,
        test_full,
        reference_empty,
        reference_full,
    ) = confusion_matrix.get_existence()

    if reference_full:
        if nan_for_nonexisting:
            return float("NaN")
        else:
            return 0.0

    return float(tn / (tn + fp))


def sensitivity(
    test=None,
    reference=None,
    confusion_matrix=None,
    nan_for_nonexisting=True,
    **kwargs
):
    """TP / (TP + FN)"""

    if confusion_matrix is None:
        confusion_matrix = ConfusionMatrix(test, reference)

    tp, fp, tn, fn = confusion_matrix.get_matrix()
    (
        test_empty,
        test_full,
        reference_empty,
        reference_full,
    ) = confusion_matrix.get_existence()

    if reference_empty:
        if nan_for_nonexisting:
            return float("NaN")
        else:
            return 0.0

    return float(tp / (tp + fn))


def dice(
    test=None,
    reference=None,
    confusion_matrix=None,
    nan_for_nonexisting=True,
    **kwargs
):
    """2TP / (2TP + FP + FN)"""

    if confusion_matrix is None:
        confusion_matrix = ConfusionMatrix(test, reference)

    tp, fp, tn, fn = confusion_matrix.get_matrix()
    (
        test_empty,
        test_full,
        reference_empty,
        reference_full,
    ) = confusion_matrix.get_existence()

    if test_empty and reference_empty:
        if nan_for_nonexisting:
            return float("NaN")
        else:
            return 0.0

    return float(2.0 * tp / (2 * tp + fp + fn))


def jaccard(
    test=None,
    reference=None,
    confusion_matrix=None,
    nan_for_nonexisting=True,
    **kwargs
):
    """TP / (TP + FP + FN)"""

    if confusion_matrix is None:
        confusion_matrix = ConfusionMatrix(test, reference)

    tp, fp, tn, fn = confusion_matrix.get_matrix()
    (
        test_empty,
        test_full,
        reference_empty,
        reference_full,
    ) = confusion_matrix.get_existence()

    if test_empty and reference_empty:
        if nan_for_nonexisting:
            return float("NaN")
        else:
            return 0.0

    return float(tp / (tp + fp + fn))


# Borrowed from https://github.com/MIC-DKFZ/nnUNet/blob/HEAD/nnunet/evaluation/metrics.py#L141-L175
def precision(
    test=None,
    reference=None,
    confusion_matrix=None,
    nan_for_nonexisting=True,
    **kwargs
):
    """TP / (TP + FP)"""

    if confusion_matrix is None:
        confusion_matrix = ConfusionMatrix(test, reference)

    tp, fp, tn, fn = confusion_matrix.get_matrix()
    (
        test_empty,
        test_full,
        reference_empty,
        reference_full,
    ) = confusion_matrix.get_existence()

    if test_empty:
        if nan_for_nonexisting:
            return float("NaN")
        else:
            return 0.0

    return float(tp / (tp + fp))


# Borrowed from https://github.com/MIC-DKFZ/nnUNet/blob/HEAD/nnunet/evaluation/metrics.py#L141-L175
def hausdorff_distance_95(
    test=None,
    reference=None,
    confusion_matrix=None,
    nan_for_nonexisting=True,
    voxel_spacing=None,
    connectivity=1,
    **kwargs
):

    if confusion_matrix is None:
        confusion_matrix = ConfusionMatrix(test, reference)

    (
        test_empty,
        test_full,
        reference_empty,
        reference_full,
    ) = confusion_matrix.get_existence()

    if test_empty or test_full or reference_empty or reference_full:
        if nan_for_nonexisting:
            return float("NaN")
        else:
            return 0

    test, reference = confusion_matrix.test, confusion_matrix.reference

    return metric.hd95(test, reference, voxel_spacing, connectivity)


# Borrowed from https://github.com/MIC-DKFZ/nnUNet/blob/HEAD/nnunet/evaluation/metrics.py#L141-L175
def avg_surface_distance(
    test=None,
    reference=None,
    confusion_matrix=None,
    nan_for_nonexisting=True,
    voxel_spacing=None,
    connectivity=1,
    **kwargs
):

    if confusion_matrix is None:
        confusion_matrix = ConfusionMatrix(test, reference)

    (
        test_empty,
        test_full,
        reference_empty,
        reference_full,
    ) = confusion_matrix.get_existence()

    if test_empty or test_full or reference_empty or reference_full:
        if nan_for_nonexisting:
            return float("NaN")
        else:
            return 0

    test, reference = confusion_matrix.test, confusion_matrix.reference

    return metric.asd(test, reference, voxel_spacing, connectivity)


def calculate_single_metric(metric_name: str, y_true, y_pred):
    y_true_tensor = torch.unsqueeze(
        torch.unsqueeze(torch.tensor(y_true), 0), 0
    )
    y_pred_tensor = torch.unsqueeze(
        torch.unsqueeze(torch.tensor(y_pred), 0), 0
    )
    # include_background - whether to skip first channel
    confusion_matrix = monai.metrics.get_confusion_matrix(
        y_pred_tensor, y_true_tensor, include_background=False
    )
    result_tensor = monai.metrics.compute_confusion_matrix_metric(
        metric_name, confusion_matrix
    )
    return result_tensor.item()


def calculate_metrics(metric_names: list, y_true, y_pred):
    result_metrics = []
    for metric_name in metric_names:
        result_metrics.append(
            calculate_single_metric(metric_name, y_true, y_pred)
        )

    return result_metrics


def calculate_metrics_bootstrap(metric_names: list, y_true, y_pred):
    result_metrics = []
    for metric_name in metric_names:
        stat, lower, upper = statistics.bootstrap_statistic(
            functools.partial(calculate_single_metric, metric_name),
            y_true,
            y_pred,
        )
        result_metrics.append((stat, lower, upper))
    return result_metrics



from typing import Sequence

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from sklearn.metrics import roc_auc_score, roc_curve

from .matplotlib_utils import get_subplots_dimensions


def hide_labels(fig, width=800, height=800):
    fig.update_layout(
        coloraxis_showscale=False, margin=dict(l=0, r=0, b=0, t=0)
    )
    fig.update_xaxes(showticklabels=False)
    fig.update_yaxes(showticklabels=False)
    fig.update_layout(width=width, height=height)
    fig.update_layout(font={"size": 40})


def boxplot_by_class(
    X: pd.DataFrame,
    y: pd.Series,
):
    """
    Plot the distributions of the selected features by the label class.
    """
    features = X.columns
    nrows, ncols, _ = get_subplots_dimensions(len(features))
    fig = make_subplots(rows=nrows, cols=ncols)
    for i, feature in enumerate(features):
        name = feature.replace("_", " ")
        fig.add_trace(
            go.Box(x=y, y=X[feature], name=name),
            row=i // ncols + 1,
            col=i % ncols + 1,
        )
    fig.update_layout(width=1300, height=700)
    fig.update_layout(font={"size": 19})
    fig.update_layout(
        legend=dict(
            orientation="h", yanchor="bottom", y=1.03, xanchor="right", x=1
        )
    )
    return fig


def plot_roc_curve(
    y_true,
    y_pred_proba,
    figsize: Sequence[int] = (600, 600),
):
    fpr, tpr, _ = roc_curve(y_true, y_pred_proba)
    roc_auc = roc_auc_score(y_true, y_pred_proba)
    fig = go.Figure(
        data=[
            go.Scatter(
                x=fpr,
                y=tpr,
                mode="lines",
                line=dict(color="royalblue", width=2),
                name=f"AUC={roc_auc:.2f}",
            )
        ],
        layout=go.Layout(
            xaxis=dict(title="False Positive Rate", showgrid=True),
            yaxis=dict(title="True Positive Rate", showgrid=False),
            font=dict(size=20),
            plot_bgcolor="rgba(0, 0, 0, 0)",
            width=figsize[0],
            height=figsize[1],
            legend=dict(yanchor="bottom", y=0.01, xanchor="right", x=0.99),
            showlegend=True,
        ),
    )
    fig.add_shape(type="line", line=dict(dash="dash"), x0=0, x1=1, y0=0, y1=1)
    fig.update_xaxes(showline=True, linewidth=2, linecolor="black")
    fig.update_yaxes(showline=True, linewidth=2, linecolor="black")
    return fig


def plot_waterfall(
    y_true: Sequence[int],
    y_pred_proba: Sequence[float],
    threshold: float,
    labels: Sequence[str] = ("positive", "negative"),
):
    y_proba_rel_to_thr = [(val - threshold) for val in y_pred_proba]
    df = (
        pd.DataFrame({"y_true": y_true, "y_pred": y_proba_rel_to_thr})
        .sort_values(by=["y_pred", "y_true"])
        .replace({"y_true": {0: labels[0], 1: labels[1]}})
    )
    fig = px.bar(
        df,
        x=range(len(y_true)),  # "plotly_index",
        y="y_pred",
        color="y_true",
        color_discrete_sequence=("red", "green"),
        labels={
            "y_pred": "Predicted score relative to threshold",
            "y_true": "",
        },
        category_orders={"y_true": labels},
        width=1000,
        height=600,
    )

    fig.update_layout(
        bargap=0.0,
        bargroupgap=0.0,
        legend=dict(yanchor="top", y=0.99, xanchor="left", x=0.01),
        font=dict(size=16),
    )
    return fig



from typing import Callable, Optional

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC
from xgboost import XGBClassifier

from autorad.config import config
from autorad.training import optuna_params


class MLClassifier(ClassifierMixin):
    """
    Class to provide a homogenous interface for all models.
    """

    def __init__(
        self,
        model,
        name: str,
        params: dict = {},
    ):
        self.model = model
        self.name = name
        self.params = params
        self._param_fn = None
        self.available_models = [
            "Random Forest",
            "AdaBoost",
            "Logistic Regression",
            "SVM",
            "Gaussian Process Classifier",
            "XGBoost",
        ]

    def __repr__(self):
        return f"{self.name}"

    @classmethod
    def from_sklearn(cls, name: str, params: dict = {}):
        if name == "Random Forest":
            model = RandomForestClassifier(**params)
        elif name == "AdaBoost":
            model = AdaBoostClassifier(**params)
        elif name == "Logistic Regression":
            model = LogisticRegression(
                max_iter=1000,
                **params,
            )
        elif name == "SVM":
            model = SVC(
                probability=True,
                max_iter=1000,
                **params,
            )
        elif name == "Gaussian Process Classifier":
            model = GaussianProcessClassifier(**params)
        elif name == "XGBoost":
            model = XGBClassifier(
                verbosity=0,
                silent=True,
                use_label_encoder=False,
                **params,
            )
        else:
            raise ValueError("Classifier name not recognized.")

        return cls(model, name, params)

    @classmethod
    def from_keras(cls, model, name, **params):
        """
        Args:
            keras_model (scikeras.KerasClassifier):
                keras model, wrapped for sklearn
            name (str): name, used to reference the model
        """
        return cls(model, name, **params)

    @classmethod
    def initialize_default_sklearn_models(cls):
        """
        Initialize a list of all available models.
        """
        models = []
        for model_name in config.AVAILABLE_CLASSIFIERS:
            model = cls.from_sklearn(model_name)
            models.append(model)
        return models

    def fit(self, X, y, **params):
        self.model.fit(X, y, **params)

    def predict(self, X):
        return self.model.predict(X)

    def predict_proba(self, X, **params):
        return self.model.predict_proba(X, **params)

    def predict_proba_binary(self, X, **params):
        return self.model.predict_proba(X, **params)[:, 1]

    def predict_label_and_proba(self, X):
        y_pred = self.model.predict(X)
        y_pred_proba = self.model.predict_proba_binary(X)

        return y_pred, y_pred_proba

    def score(self, X, y):
        return self.model.score(X, y)

    def get_params(self, deep):
        return self.model.get_params(deep)

    def set_params(self, **params):
        if self.name == "Logistic Regression":
            if params["penalty"] == "l1":
                params["solver"] = "saga"
        self.model.set_params(**params)
        self.params = params
        return self

    @property
    def param_fn(self):
        if self._param_fn is None:
            self._param_fn = optuna_params.get_param_fn(self.name)
        return self._param_fn

    # @param_fn.setter
    # def param_fn(self, param_fn: Optional[Callable] = None):
    #     if param_fn is None:
    #     self._param_fn = param_fn

    def feature_importance(self):
        if self.name == "Logistic Regression":
            importance = self.model.coef_[0]
        elif self.name in ["AdaBoost", "Random Forest", "XGBoost"]:
            importance = self.model.feature_importances_
        else:
            raise ValueError(
                f"For model {self.name} feature \
                               importance could not be calculated."
            )
        return importance


class EnsembleClassifier(BaseEstimator, ClassifierMixin):
    """
    Ensemble model for MLClassifiers.

    Parameters:
    clf : `iterable`
      A list of model objects.
    weights : `list` (default: `None`)
      If `None`, the majority rule voting will be applied to the predicted
        class labels. If a list of weights (`float` or `int`) is provided,
        the averaged raw probabilities (via `predict_proba`) will be used
        to determine the most confident class label.
    """

    def __init__(self, model_list, weights=None):
        self.model_list = model_list
        self.weights = weights

    def fit(self, X, y, **fit_params):
        """
        Fit the scikit-learn estimators.

        Parameters:
        X : numpy array, shape = [n_samples, n_features]
            Training data
        y : list or numpy array, shape = [n_samples]
            Class labels
        """
        for clf in self.model_list:
            clf.fit(X, y, **fit_params)

    def predict(self, X):
        """
        Parameters:
        X : numpy array, shape = [n_samples, n_features]

        Returns:
        maj : list or numpy array, shape = [n_samples]
            Predicted class labels by majority rule
        """
        self.classes_ = np.asarray([clf.predict(X) for clf in self.model_list])
        # if self.weights:
        #    avg = self.predict_proba(X)
        #         maj = np.apply_along_axis(lambda x: max(enumerate(x),
        #               key=operator.itemgetter(1))[0], axis=1, arr=avg)
        # else:
        #     maj = np.asarray(
        #         [
        #             np.argmax(np.bincount(self.classes_[:, c]))
        #             for c in range(self.classes_.shape[1])
        #         ]
        #     )

        # return maj

    def predict_proba(self, X):
        """
        Parameters:
        X : numpy array, shape = [n_samples, n_features]

        Returns:
        avg : list or numpy array, shape = [n_samples, n_probabilities]
            Weighted average probability for each class per sample.
        """
        self.probas_ = [clf.predict_proba(X) for clf in self.model_list]
        avg = np.average(self.probas_, axis=0, weights=self.weights)

        return avg

    def get_model_list(self):
        return self.model_list



from __future__ import annotations

import abc
import logging
import warnings
from typing import Sequence

import numpy as np
import pandas as pd
from boruta import BorutaPy
from sklearn.ensemble import RandomForestClassifier
from sklearn.exceptions import ConvergenceWarning
from sklearn.feature_selection import SelectKBest, f_classif
from sklearn.linear_model import Lasso
from sklearn.model_selection import GridSearchCV

from autorad.config import config

warnings.filterwarnings(action="ignore", category=ConvergenceWarning)
log = logging.getLogger(__name__)


class NoFeaturesSelectedError(Exception):
    """raised when feature selection fails"""

    pass


class CoreSelector(abc.ABC):
    """Template for feature selection methods"""

    def __init__(self):
        self.selected_columns = None

    @abc.abstractmethod
    def fit(self, X: np.ndarray, y: np.ndarray) -> list[int]:
        """fit method should update self.selected_columns.
        If no features are selected, it should raise
        NoFeaturesSelectedError.
        """
        pass

    def fit_transform(
        self, X: np.ndarray, y: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        self.fit(X, y)
        return X[:, self.selected_columns], y

    def transform(self, X: np.ndarray, y: np.ndarray = None) -> np.ndarray:
        if self.selected_columns is None:
            raise NoFeaturesSelectedError(
                "No features selected!" "Call fit() first before transforming."
            )
        return X[:, self.selected_columns]

    def selected_features(self, column_names: Sequence[str]):
        try:
            selected_features = [
                column_names[i] for i in self.selected_columns
            ]
        except NoFeaturesSelectedError as e:
            raise e

        return selected_features


class AnovaSelector(CoreSelector):
    def __init__(self, n_features: int = 10):
        self.n_features = n_features
        self.model = SelectKBest(f_classif, k=self.n_features)
        super().__init__()

    def fit(self, X, y):
        self.model.fit(X, y)
        support = self.model.get_support(indices=True)
        if support is None:
            raise NoFeaturesSelectedError("ANOVA failed to select features.")
        self.selected_columns = support.tolist()


class LassoSelector(CoreSelector):
    def __init__(self, alpha=0.002):
        self.model = Lasso(random_state=config.SEED, alpha=alpha)
        super().__init__()

    def optimize_params(self, X, y, verbose=0):
        search = GridSearchCV(
            self.model,
            {"alpha": np.logspace(-5, 1, num=100)},
            cv=5,
            scoring="neg_mean_squared_error",
            verbose=verbose,
        )
        search.fit(X, y)
        best_params = search.best_params_
        log.info(f"Best params for Lasso: {best_params}")
        self.model = self.model.set_params(**best_params)

    def fit(self, X, y):
        self.model.fit(X, y)
        coefficients = self.model.coef_
        importance = np.abs(coefficients)
        self.selected_columns = np.where(importance > 0)[0].tolist()
        if not self.selected_columns:
            raise NoFeaturesSelectedError("Lasso failed to select features.")

    def params_to_optimize(self):
        return {"alpha": np.logspace(-5, 1, num=100)}


class BorutaSelector(CoreSelector):
    def fit(self, X, y, verbose=0):
        model = BorutaPy(
            RandomForestClassifier(
                max_depth=5, n_jobs=-1, random_state=config.SEED
            ),
            n_estimators="auto",
            verbose=verbose,
            random_state=config.SEED,
        )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model.fit(X, y)
        self.selected_columns = np.where(model.support_)[0].tolist()
        if not self.selected_columns:
            raise NoFeaturesSelectedError("Boruta failed to select features.")

class FeatureSelectorFactory:
    def __init__(self):
        self.selectors = {
            "anova": AnovaSelector,
            "lasso": LassoSelector,
            "boruta": BorutaSelector,
        }

    def register_selector(self, name, selector):
        self.selectors[name] = selector

    def get_selector(self, name, *args, **kwargs):
        selector = self.selectors[name]
        if not selector:
            raise ValueError(f"Unknown feature selection ({name}).")
        return selector(*args, **kwargs)


def create_feature_selector(
    method: str = "anova",
    *args,
    **kwargs,
):
    selector = FeatureSelectorFactory().get_selector(method, *args, **kwargs)
    return selector



import base64
import math
import os
import re
import shutil
import uuid
from pathlib import Path

import jupytext
import pandas as pd
import streamlit as st

from autorad.config import config
from autorad.config.type_definitions import PathLike
from autorad.data.dataset import ImageDataset
from autorad.utils import conversion, io, spatial
from autorad.webapp import utils


def copy_images_to_nnunet(
    input_dir: PathLike,
    output_dir: PathLike,
    in_filename: str = "image.nii.gz",
):
    files = Path(input_dir).rglob(f"*{in_filename}")

    for file in files:
        output_path = Path(output_dir) / f"{file.parent.name}_0000.nii.gz"
        shutil.copy(str(file), str(output_path))


def copy_predictions_from_nnunet(
    pred_dir: PathLike,
    out_dir: PathLike,
    out_filename: str = "segmentation_auto.nii.gz",
):
    files = Path(pred_dir).rglob("*.nii.gz")

    for file in files:
        id_ = file.stem.split("_")[0]
        output_path = Path(out_dir) / id_ / out_filename
        shutil.copy(str(file), str(output_path))


def dicom_to_nifti_expander(data_dir):
    st.write(
        """
    For the later analysis, the images (and the segmentations) need to be
    in the **NIfTI format (.nii.gz)**. \\
    If you have DICOMs, convert them here:
    """
    )
    with st.expander("Convert DICOM to NIFTI"):
        dicom_dir = st.text_input(
            "Choose the root directory with DICOMs:",
            data_dir,
        )
        dicom_dir = Path(dicom_dir)
        if dicom_dir.is_dir():
            subdirs = [d.stem for d in dicom_dir.iterdir() if d.is_dir()]
            st.success(
                f"Directory exists and contains {len(subdirs)} subfolders. Check below if they look good.",
            )
            st.json(subdirs)
        else:
            st.error(
                f"Directory {dicom_dir} does not exist",
            )
        out_dirname = st.text_input("Output directory name:", "nifti")
        out_dir = Path(data_dir) / out_dirname
        out_dir.mkdir(exist_ok=True)
        if len(os.listdir(out_dir)) > 0:
            st.warning(f"Output directory {out_dir} is not empty. ")
        multiple_images = st.checkbox(
            "I have multiple DICOM images per patient (will require manual renaming)",
            value=False,
        )
        if not multiple_images:
            out_filename = "image"
        else:
            out_filename = None
        run_conversion = st.button("Convert")
        if run_conversion:
            with st.spinner("Converting DICOMs to NIFTI..."):
                conversion.dicom_to_nifti(
                    dicom_dir, out_dir, out_filename=out_filename
                )
            st.success(f"Done! Nifties saved in {out_dir}")

        return out_dir


def leave_only_organ_segmentation(
    seg_dir: PathLike,
    organ_label: int,
    save_dir: PathLike,
):
    """Takes the segmentations from a trained nnUNet model,
    filters out only the segmentation for `organ_label`, to which label=1 is assigned.
    The other segmentations are cleared (label=0).
    """
    label_map = {organ_label: 1}
    for mask_path in Path(seg_dir).glob("*.nii.gz"):
        save_path = Path(save_dir) / mask_path.name
        spatial.relabel_mask(
            mask_path=mask_path,
            label_map=label_map,
            save_path=save_path,
            strict=False,
            set_rest_to_zero=True,
        )


def file_selector(dir_path, text):
    filenames = os.listdir(dir_path)
    selected_filename = st.selectbox(text, filenames)
    return os.path.join(dir_path, selected_filename)


def guess_idx_of_img_colname(colnames):
    for i, colname in enumerate(colnames):
        if "img" in colname or "image" in colname:
            return i
    return 0


def guess_idx_of_seg_colname(colnames):
    for i, colname in enumerate(colnames):
        if "seg" in colname or "mask" in colname:
            return i
    return 0


def guess_idx_of_id_colname(colnames):
    for i, colname in enumerate(colnames):
        if "id" in colname.lower():
            return i
    return 0


def load_path_df():
    result_dir = utils.get_result_dir()
    path_df_path = file_selector(result_dir, "Choose a CSV table with paths:")
    path_df = pd.read_csv(path_df_path)
    st.dataframe(path_df)
    col1, col2, col3 = st.columns(3)
    colnames = path_df.columns.tolist()
    with col1:
        image_col = st.selectbox(
            "Path to image",
            colnames,
            index=guess_idx_of_img_colname(colnames),
        )
    with col2:
        mask_col = st.selectbox(
            "Path to segmentation",
            colnames,
            index=guess_idx_of_seg_colname(colnames),
        )
    with col3:
        ID_options = ["None"] + colnames
        id_col = st.selectbox(
            "ID (optional)",
            ID_options,
            index=guess_idx_of_id_colname(ID_options),
        )
    path_df.dropna(subset=[image_col, mask_col], inplace=True)
    dataset = ImageDataset(
        df=path_df,
        image_colname=image_col,
        mask_colname=mask_col,
        ID_colname=id_col,
        root_dir=config.INPUT_DIR,
    )
    return dataset


def radiomics_params():
    param_dir = Path(config.PARAM_DIR)
    presets = config.PRESETS
    preset_options = list(presets.keys())
    name = st.selectbox(
        "Choose a preset with parameters for feature extraction",
        preset_options,
    )
    preset_setup = io.read_yaml(param_dir / presets[name])
    final_setup = preset_setup.copy()

    with st.expander("Manually edit the extraction parameters"):
        final_setup = select_classes(preset_setup, final_setup)
        setting = preset_setup["setting"]
        st.write(""" Filters: """)
        filter = preset_setup["imageType"]
        col1, col2, col3 = st.columns(3)
        with col1:
            st.checkbox("original", value=("Original" in filter))
        with col2:
            turn_on_log = st.checkbox(
                "Laplacian of Gaussian",
                value=("LoG" in filter),
            )
            if turn_on_log:
                sigmas = filter["LoG"]["sigma"]
                for sigma in sigmas:
                    st.number_input("sigma", value=sigma)
        with col3:
            st.checkbox("Wavelet", value=("Wavelet" in filter))
        col1, col2, col3 = st.columns(3)
        with col1:
            st.write(""" ##### Normalization: """)
            normalization = setting["normalize"]
            is_normalized = st.checkbox("Normalize", value=normalization)
            if is_normalized:
                final_setup["setting"]["normalize"] = True
            else:
                final_setup["setting"]["normalize"] = False
        with col2:
            bin_width = st.number_input(
                """Bin width:""", value=setting["binWidth"]
            )
            final_setup["setting"]["binWidth"] = bin_width
        with col3:
            label = st.number_input("Label:", value=setting["label"])
            final_setup["setting"]["label"] = int(label)
        st.write(""" #### Full parameter file: """, preset_setup)
    return preset_setup


def choose_preset():
    config_dir = os.path.dirname(config.__file__)
    param_dir = Path(config_dir) / "pyradiomics_params"
    presets = config.PRESETS
    preset_options = list(presets.keys())
    name = st.selectbox(
        "Choose a preset with parameters for feature extraction",
        preset_options,
    )
    preset_setup = io.read_yaml(param_dir / presets[name])
    return preset_setup


def update_setup_with_class(setup, class_name, class_active):
    assert "featureClass" in setup
    if class_active[class_name]:
        setup["featureClass"][class_name] = []
    else:
        setup["featureClass"].pop(class_name, None)
    return setup


def select_classes(preset_setup, final_setup, exclude_shape=False):
    st.write(""" #### Select classes: """)

    all_feature_names = config.PYRADIOMICS_FEATURE_NAMES
    all_classes = list(all_feature_names.keys())
    if exclude_shape:
        if "shape" in all_classes:
            all_classes.remove("shape")
    preset_classes = preset_setup["featureClass"]
    class_active = {}
    cols = st.columns(len(all_classes))
    for i, class_name in enumerate(all_classes):
        with cols[i]:
            class_active[class_name] = st.checkbox(
                class_name,
                value=(class_name in preset_classes),
            )
        final_setup = update_setup_with_class(
            final_setup, class_name, class_active
        )
    st.info(
        """
        If you select a class, all its features are included by default.
        You can however select only specific features (optional):
        """
    )
    cols = st.columns(len(preset_classes.keys()))
    for i, class_name in enumerate(preset_classes.keys()):
        with cols[i]:
            if class_active[class_name]:
                feature_names = st.multiselect(
                    class_name, all_feature_names[class_name]
                )
                final_setup["featureClass"][class_name] = feature_names
    return final_setup


def radiomics_params_voxelbased() -> dict:
    param_dir = Path(config.PARAM_DIR)
    presets = config.PRESETS
    preset_options = list(presets.keys())
    name = st.selectbox(
        "Choose a preset with parameters for feature extraction",
        preset_options,
    )
    preset_setup = io.read_yaml(param_dir / presets[name])
    if "shape" in preset_setup["featureClass"]:
        preset_setup["featureClass"].pop("shape", None)
    final_setup = preset_setup.copy()

    with st.expander("Manually edit the extraction parameters"):
        final_setup = select_classes(
            preset_setup, final_setup, exclude_shape=True
        )

        setting = preset_setup["setting"]
        col1, col2, col3 = st.columns(3)
        with col1:
            st.write(""" ##### Normalization: """)
            normalization = setting["normalize"]
            is_normalized = st.checkbox("Normalize", value=normalization)
            if is_normalized:
                final_setup["setting"]["normalize"] = True
            else:
                final_setup["setting"]["normalize"] = False
        with col2:
            bin_width = st.number_input(
                """Bin width:""", value=setting["binWidth"]
            )
            final_setup["setting"]["binWidth"] = bin_width
        with col3:
            label = st.number_input("Label:", value=setting["label"])
            final_setup["setting"]["label"] = int(label)
        st.write(""" #### Full parameter file: """, preset_setup)

    return final_setup


def code_header(text):
    """
    Insert section header into a jinja file, formatted as Python comment.
    Leave 2 blank lines before the header.
    """
    seperator_len = (75 - len(text)) / 2
    seperator_len_left = math.floor(seperator_len)
    seperator_len_right = math.ceil(seperator_len)
    return f"# {'-' * seperator_len_left} {text} {'-' * seperator_len_right}"


def notebook_header(text):
    """
    Insert section header into a jinja file, formatted as notebook cell.
    Leave 2 blank lines before the header.
    """
    return f"""# # {text}
"""


def to_notebook(code):
    """Converts Python code to Jupyter notebook format."""
    notebook = jupytext.reads(code, fmt="py")
    return jupytext.writes(notebook, fmt="ipynb")


# borrowed from
# https://github.com/jrieke/traingenerator/blob/main/app/utils.py
def download_button(
    object_to_download, download_filename, button_text  # , pickle_it=False
):
    """
    Generates a link to download the given object_to_download.
    """
    try:
        # some strings <-> bytes conversions necessary here
        b64 = base64.b64encode(object_to_download.encode()).decode()
    except AttributeError:
        b64 = base64.b64encode(object_to_download).decode()

    button_uuid = str(uuid.uuid4()).replace("-", "")
    button_id = re.sub(r"\d+", "", button_uuid)

    custom_css = f"""
        <style>
            #{button_id} {{
                display: inline-flex;
                align-items: center;
                justify-content: center;
                background-color: rgb(255, 255, 255);
                color: rgb(38, 39, 48);
                padding: .25rem .75rem;
                position: relative;
                text-decoration: none;
                border-radius: 4px;
                border-width: 1px;
                border-style: solid;
                border-color: rgb(230, 234, 241);
                border-image: initial;
            }}
            #{button_id}:hover {{
                border-color: rgb(246, 51, 102);
                color: rgb(246, 51, 102);
            }}
            #{button_id}:active {{
                box-shadow: none;
                background-color: rgb(246, 51, 102);
                color: white;
                }}
        </style> """

    dl_link = (
        custom_css
        + f'<a download="{download_filename}" id="{button_id}" href="data:file/txt;base64,{b64}">{button_text}</a><br><br>'
    )

    st.markdown(dl_link, unsafe_allow_html=True)



import abc
from pathlib import Path
from typing import Any

import pandas as pd
from monai.transforms import LoadImage, SaveImage


def save_image_as_nifti_for_pyradiomics(filepath: str, tmp_dir) -> str:
    """pyRadiomics requires nifti file.
    Load image in any format (dicom/nifti/..)
    and save if as nifti in a temporary directory.

    Args:
       filepath: path to the image file / DICOM directory
    """
    image, meta = LoadImage()(filepath)
    out_dir = Path(tmp_dir)
    SaveImage(output_dir=out_dir)(image, meta)

    out_fname = Path(filepath).stem.split(".")[0] + "_trans.nii.gz"
    out_path = out_dir / out_fname

    return str(out_path)


class AbstractFeatureExtractor(abc.ABC):
    """Abstract class for all the feature extractors.
    Every extractor implementation should extend this abstract class
    and implement the methods marked as abstract: execute, describe.
    """

    def __init__(self, filepath: str, tmp_dir: str, verbose: bool = False):
        """
        Args:
            filepath: path to the image file / DICOM directory
            tmp_dir: path to the temporary directory
            verbose: logging for pyradiomics
        """
        self.filepath = filepath
        self.tmp_dir = tmp_dir
        self.verbose = verbose

    @abc.abstractmethod
    def execute(self) -> pd.DataFrame:
        """
        Returns:
            Dataframe containing the extracted features
        """
        pass

    @abc.abstractmethod
    def describe(self) -> dict[str, Any]:
        """
        Returns:
            A dict containing the attributes of the feature extractor
        """
        pass



from pathlib import Path

import streamlit as st


def check_if_dir_exists(dir_path, stop_execution=True):
    if not Path(dir_path).is_dir() or not dir_path:
        st.error(f"The entered path is not a directory ({str(dir_path)})")
        if stop_execution:
            st.stop()
    else:
        st.success(f"The entered path is a directory! ({str(dir_path)})")



import logging
from pathlib import Path

import numpy as np
from monai.data import DataLoader, Dataset
from monai.transforms import AddChanneld, Compose, LoadImaged, Spacingd
from scipy.ndimage.measurements import center_of_mass

log = logging.getLogger(__name__)


class DistanceExtractor:
    def __init__(self, df, out_path, mask1_colname, mask2_colname):
        self.df = df
        self.out_path = out_path
        self.mask1_colname = mask1_colname
        self.mask2_colname = mask2_colname
        self.ds = None
        self.loader = None

    def calculate_distance(self, arr1, arr2):
        """
        Calculates Euclidean distance between centers of mass from two binary
        arrays.
        """
        center1 = np.array(center_of_mass(arr1))
        center2 = np.array(center_of_mass(arr2))
        dist = np.linalg.norm(center1 - center2)
        return dist

    def prepare_data(self):
        files = []
        self.df[self.mask1_colname].fillna("no path", inplace=True)
        self.df[self.mask2_colname].fillna("no path", inplace=True)

        for idx, row in self.df.iterrows():
            mask1_path = row[self.mask1_colname]
            mask2_path = row[self.mask2_colname]
            if not Path(mask1_path).exists():
                log.warning(f"Path does not exist [{mask1_path}]")
            elif not Path(mask2_path).exists():
                log.warning(f"Path does not exist [{mask2_path}]")
            else:
                files.append(
                    {"mask1": mask1_path, "mask2": mask2_path, "index": idx}
                )
        transforms = Compose(
            LoadImaged(keys=["mask1", "mask2"]),
            AddChanneld(keys=["mask1", "mask2"]),
            Spacingd(keys=["mask1", "mask2"], pixdim=[1, 1, 1], diagonal=True),
        )
        self.ds = Dataset(data=files, transform=transforms)
        self.loader = DataLoader(self.ds, batch_size=1)

    def run_calculation(self):
        self.df["original_distance_from_primary"] = -100
        for data in self.loader:
            arr1 = data["mask1"].numpy()[0]
            arr2 = data["mask2"].numpy()[0]
            idx = data["index"].numpy()[0]

            dist = self.calculate_distance(arr1, arr2)
            self.df.loc[idx, "original_distance_from_primary"] = dist
        return self

    def save_df(self):
        self.df.to_csv(self.out_path)

    def execute(self):
        self.prepare_data()
        self.run_calculation()
        self.save_df()



from typing import Callable

from optuna.trial import Trial


def get_param_fn(model_name) -> Callable:
    if model_name == "Random Forest":
        param_fn = params_RandomForest
    elif model_name == "XGBoost":
        param_fn = params_XGBoost
    elif model_name == "Logistic Regression":
        param_fn = params_LogReg
    elif model_name == "SVM":
        param_fn = params_SVM
    else:
        raise ValueError(
            f"Optuna parameters for {model_name} not implemented!"
        )
    return param_fn


def params_RandomForest(trial: Trial) -> dict:
    params = {
        "n_estimators": trial.suggest_int("rf_n_estimators", 50, 1000),
        "max_depth": trial.suggest_int("rf_max_depth", 2, 50),
        "max_features": trial.suggest_categorical(
            "rf_max_features", ["auto", "sqrt"]
        ),
        "min_samples_leaf": trial.suggest_int("rf_min_samples_leaf", 1, 10),
        "min_samples_split": trial.suggest_int("rf_min_samples_split", 2, 10),
        "bootstrap": trial.suggest_categorical("rf_bootstrap", [True, False]),
    }
    return params


def params_XGBoost(trial: Trial) -> dict:
    params = {
        "lambda": trial.suggest_loguniform("xgb_lambda", 1e-8, 10.0),
        "alpha": trial.suggest_loguniform("xgb_alpha", 1e-8, 10.0),
        "colsample_bytree": trial.suggest_float("colsample_bytree", 0.2, 1.0),
        "subsample": trial.suggest_float("xgb_subsample", 0.2, 1.0),
        "booster": trial.suggest_categorical(
            "xgb_booster", ["gbtree", "gblinear", "dart"]
        ),
    }
    if params["booster"] in ["gbtree", "dart"]:
        # maximum depth of the tree, signifies complexity of the tree.
        params["max_depth"] = trial.suggest_int("xgb_max_depth", 3, 9, step=2)
        # minimum child weight, larger the term more conservative the tree.
        params["min_child_weight"] = trial.suggest_int(
            "xgb_min_child_weight", 2, 10
        )
        params["eta"] = trial.suggest_float("xgb_eta", 1e-8, 1.0, log=True)
        # defines how selective algorithm is.
        params["gamma"] = trial.suggest_float("xgb_gamma", 1e-8, 1.0, log=True)
        params["grow_policy"] = trial.suggest_categorical(
            "xgb_grow_policy", ["depthwise", "lossguide"]
        )

    if params["booster"] == "dart":
        params["sample_type"] = trial.suggest_categorical(
            "xgb_sample_type", ["uniform", "weighted"]
        )
        params["normalize_type"] = trial.suggest_categorical(
            "xgb_normalize_type", ["tree", "forest"]
        )
        params["rate_drop"] = trial.suggest_float(
            "xgb_rate_drop", 1e-8, 1.0, log=True
        )
        params["skip_drop"] = trial.suggest_float(
            "xgb_skip_drop", 1e-8, 1.0, log=True
        )
    return params


def params_SVM(trial: Trial) -> dict:
    params = {
        "kernel": trial.suggest_categorical(
            "svm_kernel", ["linear", "poly", "rbf", "sigmoid"]
        ),
        "C": trial.suggest_loguniform("svm_C", 1e-3, 10.0),
        "gamma": trial.suggest_loguniform("svm_gamma", 1e-3, 10.0),
        "degree": trial.suggest_discrete_uniform("svm_degree", 1, 5, 1),
    }
    return params


def params_LogReg(trial: Trial) -> dict:
    penalty = trial.suggest_categorical("lr_penalty", ["l2", "l1"])
    if penalty == "l1":
        solver = "saga"
    else:
        solver = "lbfgs"
    params = {
        "penalty": penalty,
        "C": trial.suggest_loguniform("lr_C", 1e-3, 10.0),
        "solver": solver,
    }
    return params


def params_preprocessing(trial):
    params = {
        "oversampling_method": trial.suggest_categorical(
            "oversampling_method",
            ["placeholder"],
        ),
    }
    return params



import functools
import logging
import warnings
from pathlib import Path
from typing import Optional

import monai.transforms.utils as monai_utils
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
import skimage
from autorad.config.type_definitions import PathLike
from autorad.utils import io, spatial
from autorad.visualization import plotly_utils
from sklearn.pipeline import Pipeline

# suppress skimage
warnings.filterwarnings(action="ignore", category=UserWarning)

log = logging.getLogger(__name__)


class Cropper:
    """Performs non-zero cropping"""

    def __init__(self, bbox_size=50, margin=20):
        self.bbox_size = bbox_size
        self.margin = margin
        self.coords_start = None
        self.coords_end = None

    def fit(self, X: np.ndarray, y=None, constant_bbox=False):
        """X is a binary mask"""
        expanded_mask = np.expand_dims(X, axis=0)
        if constant_bbox:
            select_fn = functools.partial(
                spatial.generate_bbox_around_mask_center,
                bbox_size=self.bbox_size,
            )
        else:
            select_fn = monai_utils.is_positive
        (
            self.coords_start,
            self.coords_end,
        ) = monai_utils.generate_spatial_bounding_box(
            img=expanded_mask,
            select_fn=select_fn,
            margin=[self.margin, self.margin, self.margin],
        )
        return self

    def transform(self, volume: np.ndarray):
        log.info(
            f"Cropping the image of size {volume.shape} to the region from \
            {self.coords_start} to {self.coords_end}"
        )
        return spatial.crop_volume_from_coords(
            self.coords_start, self.coords_end, volume
        )


class Slicer:
    """Given a 3D volume, finds its largest cross-section"""

    def __init__(self, axis=2):
        self.slicenum = None
        self.axis = axis

    def fit(self, X: np.ndarray, y=None):
        """X is a binary mask"""
        slicenum = spatial.get_largest_cross_section(X, axis=self.axis)
        self.slicenum = slicenum
        return self

    def transform(self, volume: np.ndarray):
        indices = [slice(None)] * 3
        indices[self.axis] = self.slicenum

        return volume[tuple(indices)]


def normalize_roi(image_array, mask_array):
    image_values = image_array[mask_array > 0]
    roi_max = np.max(image_values)
    roi_min = np.min(image_values)
    image_clipped = np.clip(image_array, roi_min, roi_max)
    image_norm = (image_clipped - roi_min) / (roi_max - roi_min)
    return image_norm


def overlay_mask_contour(
    image_2D: np.ndarray,
    mask_2D: np.ndarray,
    label: int = 1,
    color=(1, 0, 0),  # red
    normalize=False,
):
    mask_to_plot = mask_2D == label
    if normalize:
        image_2D = normalize_roi(image_2D, mask_to_plot)
    image_to_plot = skimage.img_as_ubyte(image_2D)
    result_image = skimage.segmentation.mark_boundaries(
        image_to_plot, mask_to_plot, mode="outer", color=color
    )
    return result_image


def get_plotly_fig(img):
    fig = px.imshow(img)
    fig.update_layout(width=800, height=800)
    plotly_utils.hide_labels(fig)
    return fig


def plot_roi_compare_two_masks(
    image_path, manual_mask_path, auto_mask_path, axis=2
):
    manual_vols = BaseVolumes.from_nifti(
        image_path,
        manual_mask_path,
        constant_bbox=True,
        resample=True,
        axis=axis,
    )
    auto_vols = BaseVolumes.from_nifti(
        image_path,
        auto_mask_path,
        constant_bbox=True,
        resample=True,
        axis=axis,
    )
    image_2D, manual_mask_2D = manual_vols.get_slices()
    auto_mask_2D = manual_vols.crop_and_slice(auto_vols.mask)
    img_one_contour = overlay_mask_contour(
        image_2D,
        manual_mask_2D,
    )
    img_two_contours = overlay_mask_contour(
        img_one_contour,
        auto_mask_2D,
        color=(0, 0, 1),  # blue
    )
    fig = px.imshow(img_two_contours)
    fig.update_layout(width=800, height=800)
    plotly_utils.hide_labels(fig)

    return fig


def plot_roi(image_path, mask_path):
    vols = BaseVolumes.from_nifti(image_path, mask_path, window=None)
    image_2D, mask_2D = vols.get_slices()
    img = overlay_mask_contour(image_2D, mask_2D)
    fig = get_plotly_fig(img)

    return fig


class BaseVolumes:
    """Loading and processing of image and mask volumes."""

    def __init__(
        self,
        image,
        mask,
        label=1,
        constant_bbox=False,
        window="soft tissues",
        axis=2,
    ):
        self.image_raw = image
        if window is None:
            self.image = skimage.exposure.rescale_intensity(image)
        else:
            self.image = spatial.window_with_preset(
                self.image_raw, window=window
            )
        self.mask = mask == label
        self.axis = axis
        self.preprocessor = self.init_and_fit_preprocessor(constant_bbox)

    def init_and_fit_preprocessor(self, constant_bbox=False):
        preprocessor = Pipeline(
            [
                ("cropper", Cropper()),
                ("slicer", Slicer(axis=self.axis)),
            ]
        )
        preprocessor.fit(
            self.mask,
            cropper__constant_bbox=constant_bbox,
        )
        return preprocessor

    @classmethod
    def from_nifti(
        cls,
        image_path: PathLike,
        mask_path: PathLike,
        resample=False,
        *args,
        **kwargs,
    ):

        if resample:
            mask, image = spatial.load_and_resample_to_match(
                to_resample=mask_path,
                reference=image_path,
            )
        else:
            image = io.load_image(image_path)
            mask = io.load_image(mask_path)

        return cls(image, mask, *args, **kwargs)

    def crop_and_slice(self, volume: np.ndarray):
        result = self.preprocessor.transform(volume)
        return result

    def get_slices(self):
        image_2D = self.crop_and_slice(self.image)
        mask_2D = self.crop_and_slice(self.mask)
        return image_2D.T, mask_2D.T

    def plot_image(self):
        image_2D, _ = self.get_slices()
        fig = px.imshow(image_2D, color_continuous_scale="gray")
        plotly_utils.hide_labels(fig)
        return fig


class FeaturePlotter:
    """Plotting of voxel-based radiomics features."""

    def __init__(self, image_path, mask_path, feature_map: dict):
        self.volumes = BaseVolumes.from_nifti(
            image_path, mask_path, constant_bbox=True
        )
        self.feature_map = feature_map
        self.feature_names = list(feature_map.keys())

    @classmethod
    def from_dir(cls, dir_path: str, feature_names: list[str]):
        dir_path_obj = Path(dir_path)
        image_path = dir_path_obj / "image.nii.gz"
        mask_path = dir_path_obj / "segmentation.nii.gz"
        feature_map = {}
        for name in feature_names:
            nifti_path = dir_path_obj / f"{name}.nii.gz"
            try:
                feature_map[name], _ = spatial.load_and_resample_to_match(
                    nifti_path, image_path, interpolation="bilinear"
                )
            except FileNotFoundError:
                raise FileNotFoundError(
                    f"Could not load feature map {name} in {dir_path}"
                )
        return cls(image_path, mask_path, feature_map)

    def plot_single_feature(
        self,
        feature_name: str,
        feature_range: Optional[tuple[float, float]] = None,
    ):
        image_2D, mask_2D = self.volumes.get_slices()
        feature_2D = self.volumes.crop_and_slice(
            self.feature_map[feature_name]
        )
        feature_2D[mask_2D == 0] = np.nan
        fig = px.imshow(image_2D, color_continuous_scale="gray")
        plotly_utils.hide_labels(fig)
        # Plot mask on top of fig, without margins
        heatmap_options = {
            "z": feature_2D,
            "showscale": False,
            "colorscale": "Spectral",
        }
        if feature_range:
            heatmap_options["zmin"] = feature_range[0]
            heatmap_options["zmax"] = feature_range[1]
        fig.add_trace(go.Heatmap(**heatmap_options))
        return fig

    def plot_all_features(self, output_dir, param_ranges):
        fig = self.volumes.plot_image()
        fig.write_image(Path(output_dir) / "image.png")
        for name, param_range in zip(self.feature_names, param_ranges):
            fig = self.plot_single_feature(name, param_range)
            fig.write_image(Path(output_dir) / f"{name}.png")



import logging
import pickle
from pathlib import Path
from typing import Sequence

import mlflow
import numpy as np
import optuna
from optuna.integration.mlflow import MLflowCallback
from optuna.study import Study
from optuna.trial import Trial
from sklearn.metrics import roc_auc_score

from autorad.config import config
from autorad.config.type_definitions import PathLike
from autorad.data.dataset import FeatureDataset
from autorad.models.classifier import MLClassifier
from autorad.preprocessing.preprocessor import Preprocessor
from autorad.training import utils
from autorad.training.optimizer import OptunaOptimizer
from autorad.utils import io

log = logging.getLogger(__name__)


class Trainer:
    """
    Runs the experiment that optimizes the hyperparameters
    for all the models, given the dataset with extracted features.
    """

    def __init__(
        self,
        dataset: FeatureDataset,
        models: Sequence[MLClassifier],
        result_dir: PathLike,
        experiment_name: str = "baseline",
    ):
        self.dataset = dataset
        self.models = models
        self.result_dir = Path(result_dir)
        self.experiment_name = experiment_name

        self.registry_dir = self.result_dir / "models"
        self._optimizer = None
        self.auto_preprocessing = False
        utils.init_mlflow(self.registry_dir)

    def set_optimizer(self, optimizer: str, n_trials=100):
        if optimizer == "optuna":
            self._optimizer = OptunaOptimizer(n_trials=n_trials)
        # elif optimizer == "gridsearch":
        #     self.optimizer = GridSearchOptimizer()
        else:
            raise ValueError("Optimizer not recognized.")

    def run_auto_preprocessing(
        self, oversampling=True, selection_methods=None
    ):
        if selection_methods is None:
            selection_methods = config.FEATURE_SELECTION_METHODS
        if oversampling:
            oversampling_methods = config.OVERSAMPLING_METHODS
        else:
            oversampling_methods = [None]
        preprocessed = {}
        for selection_method in selection_methods:
            preprocessed[selection_method] = {}
            for oversampling_method in oversampling_methods:
                preprocessor = Preprocessor(
                    normalize=True,
                    feature_selection_method=selection_method,
                    oversampling_method=oversampling_method,
                )
                try:
                    preprocessed[selection_method][
                        oversampling_method
                    ] = preprocessor.fit_transform(self.dataset.data)
                except AssertionError:
                    log.error(
                        f"Preprocessing with {selection_method} and {oversampling_method} failed."
                    )
            if not preprocessed[selection_method]:
                del preprocessed[selection_method]
        with open(self.result_dir / "preprocessed.pkl", "wb") as f:
            pickle.dump(preprocessed, f, pickle.HIGHEST_PROTOCOL)

    @property
    def optimizer(self):
        if self._optimizer is None:
            raise ValueError("Optimizer is not set!")
        return self._optimizer

    def set_optuna_params(self, model: MLClassifier, trial: Trial):
        params = model.param_fn(trial)
        model.set_params(**params)
        return model

    def run(
        self,
        auto_preprocess: bool = False,
    ):
        """
        Run hyperparameter optimization for all the models.
        """
        mlfc = MLflowCallback(
            tracking_uri=mlflow.get_tracking_uri(), metric_name="AUC"
        )
        study = self.optimizer.create_study(study_name=self.experiment_name)
        study.optimize(
            lambda trial: self._objective(trial, auto_preprocess),
            n_trials=self.optimizer.n_trials,
            callbacks=[mlfc],
        )
        self.save_best_params(study)

    def save_best_params(self, study: Study):
        params = study.best_trial.params
        io.save_json(params, (self.result_dir / "best_params.json"))

    def optimize_preprocessing(self, trial: Trial):
        pkl_path = self.result_dir / "preprocessed.pkl"
        with open(pkl_path, "rb") as f:
            preprocessed = pickle.load(f)
        feature_selection_method = trial.suggest_categorical(
            "feature_selection_method", preprocessed.keys()
        )
        oversampling_method = trial.suggest_categorical(
            "oversampling_method",
            preprocessed[feature_selection_method].keys(),
        )
        result = preprocessed[feature_selection_method][oversampling_method]

        return result

    def _objective(self, trial: optuna.Trial, auto_preprocess=False) -> float:
        """Get params from optuna trial, return the metric."""
        if auto_preprocess:
            data = self.optimize_preprocessing(trial)
        else:
            data = self.dataset.data

        model_name = trial.suggest_categorical(
            "model", [m.name for m in self.models]
        )
        model = utils.get_model_by_name(model_name, self.models)
        model = self.set_optuna_params(model, trial)
        aucs = []
        for (
            X_train,
            y_train,
            _,
            X_val,
            y_val,
            _,
        ) in data.iter_training():
            model.fit(X_train, y_train)
            y_pred = model.predict_proba_binary(X_val)
            auc_val = roc_auc_score(y_val, y_pred)
            aucs.append(auc_val)
        AUC = np.mean(aucs)

        return AUC



import logging
from typing import Callable, Sequence

import numpy as np

from autorad.utils import io

log = logging.getLogger(__name__)


def assert_equal_shape(arr1, arr2):
    assert (
        arr1.shape == arr2.shape
    ), f"Shape of arrays {arr1.shape} and {arr2.shape} are not equal"


def assert_has_nonzero(arr):
    assert (arr == 1).sum() > 0, "Array has only zeros"


def assert_has_nonzero_within_roi(arr, mask):
    assert (
        arr[mask == 1]
    ).sum() > 0, "Array has only zeros within the region where mask==1"


def assert_not_equal(arr1, arr2):
    assert not np.array_equal(arr1, arr2), "Arrays are the same"


def assert_no_empty_slice_in_mask(array_3D):
    assert (array_3D.sum(axis=(1, 2)) > 0).all(), "Array has empty slices"


def check_assertion_from_paths(assert_fn, paths: Sequence[str]):
    args = [io.load_image(path) for path in paths]
    try:
        assert_fn(*args)
        return None
    except AssertionError as e:
        msg = f"{e}, file: {paths} \n "
        return msg


def pack_as_list(arg):
    if isinstance(arg, str):
        result = [arg]
    else:
        result = list(arg)
    return result


def check_assertion_dataset(
    assert_fn: Callable, paths: Sequence[str | tuple[str, str]]
):

    asserts = [
        check_assertion_from_paths(assert_fn, pack_as_list(path))
        for path in paths
    ]
    failed_asserts = [a for a in asserts if a is not None]
    if len(failed_asserts) > 0:
        raise AssertionError(" ".join(failed_asserts))



from pathlib import Path

import streamlit as st

from autorad.utils import preprocessing, testing
from autorad.webapp import template_utils, utils


def show():
    st.write(
        "Once you've created the table from previous step, run some basic tests on the images and masks:"
    )
    dataset = template_utils.load_path_df()
    n_cases = len(dataset.df)
    st.write("##### Test your data:")
    if st.button("Test if ROI has at least one pixel"):
        try:
            testing.check_assertion_dataset(
                testing.assert_has_nonzero, dataset.mask_paths
            )
        except AssertionError as e:
            st.error(e)
        else:
            st.success(f"Passed for all {n_cases} cases!")
    if st.button("Test if image and masks are of the same size"):
        try:
            testing.check_assertion_dataset(
                testing.assert_equal_shape,
                list(zip(dataset.image_paths, dataset.mask_paths)),
            )
        except AssertionError as e:
            st.error(e)
        else:
            st.success(f"Passed for {n_cases} cases!")
    if st.button("Test if ROI in the image has >1 unique values"):
        try:
            testing.check_assertion_dataset(
                testing.assert_has_nonzero_within_roi,
                list(zip(dataset.image_paths, dataset.mask_paths)),
            )
        except AssertionError as e:
            st.error(e)
        else:
            st.success(f"Passed for all {n_cases} cases!")
    st.write(
        """
    #####
    You can proceed to the next step, [**Feature extraction**](#feature-extraction). \n
    Or, if you want to add additional **peritumoral masks** of the region surrounding the segmentation, you can do it below ⬇
    """,
        unsafe_allow_html=True,
    )
    col1, col2 = st.columns(2)
    with col1:
        margin = st.number_input(label="Margin (in mm)", value=5)
        margin = int(margin)
    input_dir = Path(utils.get_input_dir())
    result_dir = Path(utils.get_result_dir())
    save_dir = input_dir / "peritumoral_masks"
    with col2:
        st.text_input(
            "Where to save the peritumoral masks",
            value=str(save_dir),
        )
    if st.button("Generate peritumoral masks"):
        with st.spinner("Generating border masks..."):
            result_dir.mkdir(exist_ok=True)
            dilated_mask_dir = save_dir / f"border_masks_{margin}"
            dilated_mask_dir.mkdir(exist_ok=True, parents=True)
            extended_paths_df = preprocessing.generate_border_masks(
                dataset=dataset,
                margin_in_mm=int(margin),
                output_dir=dilated_mask_dir,
            )
        st.dataframe(extended_paths_df)
        utils.save_table_in_result_dir(
            extended_paths_df,
            "paths_with_border_masks.csv",
            button=False,
        )


if __name__ == "__main__":
    show()



import logging

import numpy as np
from sklearn.metrics import roc_auc_score

from autorad.data.dataset import FeatureDataset, TrainingData
from autorad.models.classifier import MLClassifier
from autorad.preprocessing.preprocessor import Preprocessor
from autorad.utils import io

log = logging.getLogger(__name__)


class Inferrer:
    def __init__(self, params, result_dir):
        self.params = params
        self.result_dir = result_dir
        self.model, self.preprocessor = self._parse_params()

    def _parse_params(self):
        temp_params = self.params.copy()
        selection = temp_params.pop("feature_selection_method")
        oversampling = temp_params.pop("oversampling_method")
        preprocessor = Preprocessor(
            normalize=True,
            feature_selection_method=selection,
            oversampling_method=oversampling,
        )
        model = MLClassifier.from_sklearn(temp_params.pop("model"))
        model_params = {
            "_".join(k.split("_")[1:]): v for k, v in temp_params.items()
        }
        model.set_params(**model_params)

        return model, preprocessor

    def fit(self, dataset: FeatureDataset):
        _data = self.preprocessor.fit_transform(dataset.data)
        self.model.fit(
            _data._X_preprocessed.train, _data._y_preprocessed.train
        )

    def eval(self, dataset: FeatureDataset, result_name: str = "results"):
        X = self.preprocessor.transform(dataset.data.X.test)
        y = dataset.data.y.test
        y_pred = self.model.predict_proba_binary(X)
        auc = roc_auc_score(y, y_pred)
        result = {}
        result["selected_features"] = self.preprocessor.selected_features
        result["AUC test"] = auc
        io.save_json(result, (self.result_dir / f"{result_name}.json"))
        io.save_predictions_to_csv(
            y, y_pred, (self.result_dir / f"{result_name}.csv")
        )

    def fit_eval(self, dataset: FeatureDataset, result_name: str = "results"):
        _data = self.preprocessor.fit_transform(dataset.data)
        result = {}
        result["selected_features"] = _data.selected_features
        train_auc = self._fit_eval_splits(_data)
        result["AUC train"] = train_auc
        test_auc = self._fit_eval_train_test(_data, result_name)
        result["AUC test"] = test_auc
        log.info(
            f"Test AUC: {test_auc:.3f}, mean train AUC: {np.mean(train_auc):.3f}"
        )
        io.save_json(result, (self.result_dir / f"{result_name}.json"))

    def _fit_eval_train_test(
        self, _data: TrainingData, result_name: str = "results"
    ):
        self.model.fit(
            _data._X_preprocessed.train, _data._y_preprocessed.train
        )
        y_pred = self.model.predict_proba_binary(_data._X_preprocessed.test)
        y_test = _data._y_preprocessed.test
        auc = roc_auc_score(y_test, y_pred)
        io.save_predictions_to_csv(
            y_test, y_pred, (self.result_dir / f"{result_name}.csv")
        )
        return auc

    def _fit_eval_splits(self, data: TrainingData):
        aucs = []
        for X_train, y_train, _, X_val, y_val, _ in data.iter_training():
            self.model.fit(X_train, y_train)
            y_pred = self.model.predict_proba_binary(X_val)
            auc_val = roc_auc_score(y_val, y_pred)
            aucs.append(auc_val)

        return aucs

    def init_result_df(self, dataset: FeatureDataset):
        self.result_df = dataset.meta_df.copy()
        self.test_indices = dataset.X.test.index.values



import logging
import shutil
from pathlib import Path

import radiomics
import SimpleITK as sitk
from autorad.config import config
from autorad.config.type_definitions import PathLike
from autorad.utils import io


def extract_feature_maps(
    image_path: PathLike,
    seg_path: PathLike,
    save_dir: PathLike,
    extraction_params: dict | None = None,
    copy_inputs: bool = True,
):
    save_dir = Path(save_dir)
    if copy_inputs:
        shutil.copyfile(image_path, save_dir / "image.nii.gz")
        shutil.copyfile(seg_path, save_dir / "segmentation.nii.gz")
    if extraction_params is None:
        extraction_params = io.read_yaml(
            Path(config.PARAM_DIR) / "CT_default_feature_map.yaml"
        )
    radiomics.setVerbosity(logging.INFO)
    extractor = radiomics.featureextractor.RadiomicsFeatureExtractor(
        extraction_params
    )
    feature_vector = extractor.execute(
        str(image_path), str(seg_path), voxelBased=True
    )
    for feature_name, feature_value in feature_vector.items():
        if isinstance(feature_value, sitk.Image):
            save_path = save_dir / f"{feature_name}.nii.gz"
            save_path = str(save_path)
            sitk.WriteImage(feature_value, save_path)



import streamlit as st

from autorad.webapp import utils, webapp_config


def show_sidebar_and_select_template():
    template_dict = webapp_config.TEMPLATE_DICT
    task = "Binary classification"
    with st.sidebar:
        st.write("## Task")
        framework = st.radio("Select task:", list(template_dict[task].keys()))
    template_path = template_dict[task][framework]
    show_template(template_path)


def show_template(template_path):
    template = utils.import_from_file("template", template_path)
    template.show()


def show_title():
    col1, col2 = st.columns(2)
    with col1:
        st.title("AutoRadiomics")
    with col2:
        st.write(
            """
        ####
        The easiest framework for experimenting
        using `pyradiomics` and `scikit-learn`.
        """
        )


def main():
    st.set_page_config(
        page_title="AutoRadiomics",
        layout="wide",
    )
    show_title()
    show_sidebar_and_select_template()


if __name__ == "__main__":
    main()



import setuptools
from setuptools import setup


with open("requirements.txt") as file:
    required_packages = [ln.strip() for ln in file.readlines()]

test_packages = [
    "coverage==6.2",
    "great-expectations==0.14.2",
    "pytest==6.2.5",
    "pytest-watch",
    "hypothesis==6.36.0",
]

dev_packages = [
    "black==21.12b0",
    "flake8==4.0.1",
    "isort==5.10.1",
    "pre-commit==2.17.0",
]

webapp_packages = ["streamlit==1.10.0", "docker==5.0.3", "jupytext==1.14.0"]

docs_packages = [
    "mkdocs==1.3.0",
    "mkdocs-material==8.3.3",
    "mkdocstrings==0.18.1",
    "mkdocstrings-python-legacy",
]

setup(
    name="autorad",
    packages=setuptools.find_packages(),
    package_data={
        "autorad": [
            "config/pyradiomics_feature_names.json",
            "webapp/paths_example.csv",
            "webapp/templates/segmentation/pretrained_models.json",
            "config/pyradiomics_params/*",
        ]
    },
    include_package_data=True,
    install_requires=required_packages,
    extras_require={
        "app": webapp_packages,
        "dev": test_packages + dev_packages + webapp_packages + docs_packages,
        "docs": docs_packages,
    },
    entry_points={
        "console_scripts": [
            "dicom_to_nifti = autorad.utils.preprocessing:dicom_app",
            "nrrd_to_nifti = autorad.utils.preprocessing:nrrd_app",
            "utils = autorad.utils.utils:app",
        ],
    },
)



import logging
import os
from pathlib import Path
from typing import Optional

import SimpleITK as sitk
import typer
from monai.transforms import Compose, LoadImage, SaveImage
from nipype.interfaces.dcm2nii import Dcm2niix

from autorad.config.type_definitions import PathLike

log = logging.getLogger(__name__)

dicom_app = typer.Typer()
nrrd_app = typer.Typer()


def convert_dataset_to_nifti(data_dir: PathLike, save_dir: PathLike):
    """Convert from DICOM or nrrd to nifti"""
    data_dir = Path(data_dir)
    if not data_dir.exists():
        raise ValueError(f"{data_dir} does not exist")
    data = list(data_dir.glob("*"))[:-1]
    if len(data) == 0:
        raise ValueError(f"{data_dir} is empty")
    save_dir = Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)
    io_pipeline = Compose(
        [
            LoadImage(ensure_channel_first=True, reader="ITKReader"),
            SaveImage(output_dir=save_dir, output_postfix=""),
        ]
    )
    io_pipeline(data)


def get_dcm2niix_converter(
    dicom_dir: Path, save_dir: Path, out_filename: Optional[str] = None
):
    """
    Args:
        dicom_dir: directory with DICOM Study
        save_dir: directory where to save the nifti
        out_filename: what to include in filenames (from dcm2niix)
    Returns:
        converter: nipype Dcm2niix object based on rodenlab dcm2niix with basic
                   default parameters to merge Dicom series into 3d Volumes
    """
    converter = Dcm2niix()
    converter.inputs.source_dir = str(dicom_dir)
    converter.inputs.output_dir = str(save_dir)
    converter.inputs.merge_imgs = True
    if out_filename:
        converter.inputs.out_filename = out_filename

    return converter


@dicom_app.command()
def dicom_to_nifti(
    input_dir: PathLike,
    output_dir: PathLike,
    subdir_name: str = "",
    out_filename: Optional[str] = None,
):
    """
    Args:
        input_dir: absolute path to the directory with all the cases containing
                dicoms
        output_dir: absolute path to the directory where to save nifties
        subdir_name: optional name of subdirectory within case dir
    """
    for id_ in os.listdir(input_dir):
        dicom_dir = Path(input_dir) / id_ / subdir_name
        save_dir = Path(output_dir) / id_ / subdir_name
        save_dir.mkdir(exist_ok=True, parents=True)
        if not dicom_dir.exists():
            raise FileNotFoundError(f"Dicom directory {dicom_dir} not found.")
        converter = get_dcm2niix_converter(dicom_dir, save_dir, out_filename)
        converter.run()


@nrrd_app.command()
def nrrd_to_nifti(nrrd_path: str, output_path: str):
    """
    Converts an image in NRRD format to NifTI format.
    """
    img = sitk.ReadImage(nrrd_path)
    Path(output_path).parent.mkdir(exist_ok=True, parents=True)
    sitk.WriteImage(img, str(output_path))



from pathlib import Path

import seaborn as sns
import streamlit as st

from autorad.utils import io
from autorad.visualization import plot_volumes
from autorad.webapp import template_utils, utils
from autorad.webapp.extractor import StreamlitFeatureExtractor
from autorad.webapp.template_utils import radiomics_params


def show():
    st.header("Feature extraction")
    with st.sidebar:
        st.write(
            """
            Expected input:
                CSV file with with absolute paths to the image and the mask
                for each case.
        """
        )
    dataset = template_utils.load_path_df()
    result_dir = Path(utils.get_result_dir())
    result_dir.mkdir(exist_ok=True)
    with st.expander("Inspect the data"):
        if st.button("Draw random case"):
            row = dataset.df.sample(1)
            st.dataframe(row)
            image_path = row[dataset.image_colname]
            mask_path = row[dataset.mask_colname]
            try:
                fig = plot_volumes.plot_roi(image_path, mask_path)
                fig.update_layout(width=300, height=300)
                st.plotly_chart(fig)
            except TypeError:
                raise ValueError(
                    "Image or mask path is not a string. "
                    "Did you correctly set the paths above?"
                )

    extraction_params = radiomics_params()
    col1, col2 = st.columns(2)
    with col1:
        filename = st.text_input(
            "Name for the table with features", "features.csv"
        )
    with col2:
        n_jobs = st.slider(
            "Number of threads", min_value=1, max_value=8, value=1
        )
    extraction_button = st.button("Run feature extraction")
    if extraction_button:
        params_path = result_dir / "extraction_params.json"
        io.save_json(extraction_params, params_path)
        extractor = StreamlitFeatureExtractor(
            dataset=dataset,
            n_jobs=n_jobs,
            extraction_params=params_path,
        )
        with st.spinner("Extracting features"):
            feature_df = extractor.run()
        cm = sns.light_palette("green", as_cmap=True)
        display_df = feature_df.style.background_gradient(cmap=cm)
        st.dataframe(display_df)
        utils.save_table_in_result_dir(feature_df, filename, button=False)


if __name__ == "__main__":
    show()
